﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NAudio.Wave;
using System.Threading;
using System.IO;

namespace WpfAudioPlayer.Models.Wmf
{
	/// <summary>
	/// WmfReader を独自のスレッドに包み、同期的な操作を行う為のクラスです。
	/// </summary>
	class WmfSyncReader : WaveStream
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <param name="fileName">ファイルのパス。</param>
		public WmfSyncReader( string fileName )
		{
			this._status = ReaderState.Initializing;
			this._asyncOwnerThread = new AutoResetEvent( false );
			this._readerThread = new Thread( ReadThreadProc );

			this._readerThread.Start( fileName );
			this._asyncOwnerThread.WaitOne();
		}
		/// <summary>
		/// 音声データの読み取りスレッドとして実行されるプロシージャ。
		/// </summary>
		/// <param name="param">ファイルのパス。</param>
		private void ReadThreadProc( object param )
		{
			string fileLocation = param as string;

			using( this._reader = new WmfReader( fileLocation ) )
			{
				this._asyncReaderThread = new AutoResetEvent( false );
				this._status = ReaderState.Reading;
				this._asyncOwnerThread.Set();
				this._asyncReaderThread.WaitOne();

				while( this._status == ReaderState.Reading || this._status == ReaderState.Seeking )
				{
					if( this._status == ReaderState.Seeking )
					{
						// ストリーム内の位置移動
						this._status = ReaderState.Reading;
						this._seekParam.Position = this._reader.Seek( this._seekParam.Offset, this._seekParam.Origin );
					}
					else
					{
						// 音声データの読み取り
						this._buffer.Read = this._reader.Read( this._buffer.Buffer, this._buffer.Offset, this._buffer.Count );
					}

					this._asyncOwnerThread.Set();
					this._asyncReaderThread.WaitOne();
				}

				this._asyncOwnerThread.Set();
			}
		}

		/// <summary>
		/// 外部から呼び出される、音声ストリーム操作系メソッド間の排他実行を行う為のメソッド。
		/// </summary>
		private void WaitOne()
		{
			lock( this )
			{
				this._asyncReaderThread.Set();
				this._asyncOwnerThread.WaitOne();
			}
		}

		#region WaveStream の上書き

		/// <summary>
		/// リソースの解放を行います。
		/// </summary>
		/// <param name="disposing">リソースの解放を行う場合は true。それ以外は false。</param>
		protected override void Dispose( bool disposing )
		{
			if( this._reader != null )
			{
				this._status = ReaderState.Disposing;
				this.WaitOne();
				this._reader = null;
			}

			base.Dispose( disposing );
		}

		/// <summary>
		/// 現在のストリーム内の位置を設定します。
		/// </summary>
		/// <param name="offset">origin パラメータからのバイト オフセット。</param>
		/// <param name="origin">新しい位置を取得するために使用する参照ポイントを示す SeekOrigin 型の値。</param>
		/// <returns>現在のストリーム内の新しい位置。</returns>
		public override long Seek( long offset, SeekOrigin origin )
		{
			this._seekParam = new SeekParam { Offset = offset, Origin = origin };
			this._status = ReaderState.Seeking;
			this.WaitOne();
			return this._seekParam.Position;
		}

		/// <summary>
		/// シーク操作をサポートしている事を示す値を取得します。
		/// </summary>
		public override bool CanSeek
		{
			get
			{
				return this._reader.CanSeek;
			}
		}

		/// <summary>
		/// 現在の音声データ中の位置を時間単位で取得または設定する。
		/// </summary>
		public new TimeSpan CurrentTime
		{
			get
			{
				return this._reader.CurrentTime;
			}
			set
			{
				this.Position = ( long )( value.TotalSeconds * this.WaveFormat.AverageBytesPerSecond );
			}
		}


		/// <summary>
		/// 総演奏時間を取得します。
		/// </summary>
		public new TimeSpan TotalTime
		{
			get
			{
				return this._reader.TotalTime;
			}
		}

		#endregion

		#region WaveStream の実装

		/// <summary>
		/// オーディオ情報を取得します。
		/// </summary>
		public override WaveFormat WaveFormat
		{
			get
			{
				return this._reader.WaveFormat;
			}
		}

		/// <summary>
		/// オーディオ データの長さを取得します。
		/// </summary>
		public override long Length
		{
			get
			{
				return this._reader.Length;;
			}
		}

		/// <summary>
		/// オーディオ データ内の位置を取得または設定します。
		/// </summary>
		public override long Position
		{
			get
			{
				return this._reader.Position;
			}
			set
			{
				this.Seek( value, SeekOrigin.Begin );
			}
		}

		/// <summary>
		/// オーディオ データを読み取ります。
		/// </summary>
		/// <param name="buffer">読み取ったデータを格納するバッファ。</param>
		/// <param name="offset">読み取り位置。</param>
		/// <param name="count">読み取るサイズ。</param>
		/// <returns>読み取りに成功したサイズ。</returns>
		public override int Read( byte[] buffer, int offset, int count )
		{
			this._buffer = new ReadBuffer { Buffer = buffer, Offset = offset, Count = count };
			this.WaitOne();
			return this._buffer.Read;
		}

		#endregion


		#region フィールド

		/// <summary>
		/// 音声データ読み取りスレッドの状態。
		/// </summary>
		private ReaderState _status;

		/// <summary>
		/// 音声データ読み取り用スレッド。
		/// </summary>
		private Thread _readerThread;

		/// <summary>
		/// 音声データ読み取り用スレッドを同期実行する為のイベント。
		/// </summary>
		private AutoResetEvent _asyncReaderThread;

		/// <summary>
		/// このクラスを所有しているスレッドを待機させる為のイベント。
		/// </summary>
		private AutoResetEvent _asyncOwnerThread;

		/// <summary>
		/// Windows Media オーディオの読み取り専用ストリーム。
		/// </summary>
		private WmfReader _reader;

		/// <summary>
		/// 読み取られた音声データを格納する為のバッファ。
		/// </summary>
		private ReadBuffer _buffer;

		/// <summary>
		/// 現在のストリーム内の位置を設定する為のパラメータ。
		/// </summary>
		private SeekParam _seekParam;

		#endregion
		
		#region 内部クラス、列挙体

		/// <summary>
		/// 現在のインスタンスの状態を定義した列挙体。
		/// </summary>
		private enum ReaderState
		{
			/// <summary>
			/// 初期化中。
			/// </summary>
			Initializing,

			/// <summary>
			/// 音声データの読み込みを実行中。
			/// </summary>
			Reading,

			/// <summary>
			/// ストリーム内の位置を設定中。
			/// </summary>
			Seeking,

			/// <summary>
			/// リソースの解放を実行中。
			/// </summary>
			Disposing
		}

		/// <summary>
		/// 音声データを格納する為の読み取り専用バッファ。
		/// </summary>
		private class ReadBuffer
		{
			/// <summary>
			/// 読み込まれたバイト数。
			/// </summary>
			public int Read;

			/// <summary>
			/// 読み込まれた音声データを格納するバッファ。
			/// </summary>
			public byte[] Buffer;

			/// <summary>
			/// 読み取ったデータの格納を開始する位置を示す buffer 内のバイト オフセット。
			/// </summary>
			public int Offset;

			/// <summary>
			/// 現在のストリームから読み取る最大バイト数。
			/// </summary>
			public int Count;
		}

		/// <summary>
		/// 現在のストリーム内の位置を設定する為のパラメータ。
		/// </summary>
		private class SeekParam
		{
			/// <summary>
			/// ストリーム内の新しい位置。
			/// </summary>
			public long Position;

			/// <summary>
			/// Origin メンバからのバイト オフセット。
			/// </summary>
			public long Offset;

			/// <summary>
			/// 新しい位置を取得するために使用する参照ポイントを示す SeekOrigin 型の値。
			/// </summary>
			public SeekOrigin Origin;
		}

		#endregion
	}
}
